#ifndef Enumerate_h
#define Enumerate_h

#include "LPBIneqSet.h"
#include "LPBIRefs.h"
#include "LitRefs.h"
#include "SolStack.h"


class Enumerate {
public:
  Enumerate();
  Enumerate(const Enumerate&);
  Enumerate& operator=(const Enumerate&);
  ~Enumerate();

// initialization and changes
  void clear();
  int add(int iid, LPBIneq& lpbi);
  int del(int iid);
  int replace(int iid, LPBIneq& lpbi);

  int set_heuristic(int );
  int get_heuristic();
  int set_verbosity(int );
  int get_verbosity();

// sets all up for starting enumeration;  
  int init_enumeration();
  int reset_enumeration();

// consistency
  int solve();
  int solve(int fixlit);
  int solve_limit(int nodes=1000); // at most 'nodes' nodes are explored

// optimization
  // ones are all variables that are 1 in the found solution 
  // max = 1 if we maximize
  // lb = 1 if an initial bound should be computed
  Coefficient optimize(Products objf,int& consistent, OrdInt& sol, int max=1, int lb=0);
  int maximize();
  int maximize(int);

// for usage "by hand"
  void EnumTablesave();
  void EnumTablerestore();

  // <- UNSAT then cover contains a set of literals whose fixing
  // yields unsat; otherwise it contains a solution, i.e. fixing
  // of the literals in cover yields the empty problem
  int GenerateCover(OrdInt& cover);
  // the cover in the order the literals have been selected
  RArray<int> last_cover; 

  // eliminates some literals in cover but fixing the literals in cover 
  // still yields an unsatisfiable problem
  // if a solution is found; solution is updated. 
  // If lb (we are calculating a lower bound) then unsatisfiability
  //    of the cover is guaranteed and UNSAT returned
  // else SAT is returned and solution is updated
  int MinimizeCover(OrdInt& cover, int max_cut_nodes = 0, int lb=0); 
  int select_to_del_lit(OrdInt& to_del); // internal
  int select_to_del_lit_option;

  // computes a lower bound of term wrt. the current set of constraints
  Coefficient LowerBound(Products term);

  OrdInt current_fixed(); // fixed_literals of current state
  OrdInt all_fixed();     // all fixed literals in current branch

  SolStack last_solution; // stores the solution whenever one is found 
  void update_solution();

// statistics 
  void stat_reset();
  int stat_nodes();
  int stat_maxdepth();
  int stat_fixedlits();
  int compute_sizes();
  int show_enum_table();

// the return values from fixing
  enum {SAT = 0,UNSAT = -1, OBJUNSAT = -2, OPEN = 1,UNLINK = 2};

// the state of an LPBIneq
  class LPBIState {
  public:
    Coefficient sumci;
    Coefficient rhs;
    Coefficient maxci;
    int maxliteral;
    LPBIState() {}
    LPBIState(const LPBIState& l) : sumci(l.sumci), rhs(l.rhs), maxci(l.maxci), maxliteral(l.maxliteral) {}
    void operator=(const LPBIState& l) { sumci = l.sumci; rhs = l.rhs; maxci = l.maxci; maxliteral = l.maxliteral; }
    ~LPBIState() {}
  };

private:
// the stack where the fixed literals are stored
  RArray<int> FixedStack;
  int FixedStackIndex;
  RArray<int> FixedStackChoicepoints;
  int FixedStackChoicepointsIndex;

// the stack where literals that still need to be fixed are stored
  RArray<int> ToFixStack;
  int ToFixStackIndex;
  
// the current value of the literals
// -1 fixed to 0, 0 open, 1 fixed to 1
  RArray<int> LitValue;

// all products of all inequalities (static while enumerating)
  LPBIRefs lpbiref;
// the literal references
  LitRefs litref;

// the tablestack where all the current states of all ineqs are stored
// for all nodes in the current path
  RArray< RArrayList<LPBIState> > EnumTableStack;
  int EnumTableStackIndex;
  int largestenlargedETSindex; // largest EnumTableStackindex which has 
                               // already needed memory

// for the heuristics
  RArray<double> LitHeuValue;

// for optimizing
  Coefficient local_maximum; // to reach
  Coefficient last_local_maximum; // reached
  int objective_function_iid;
  int do_maximize;
  Coefficient objf_sumcoeffs;
  Coefficient objf_constant;
  int show_each_sol;

// Statistics
  int nodes;
  int fixedlits;
  float start_time;

// Verbosity
  int verbosity;
#ifdef VERBOSE
#define ENUM_VERB(i,t) if (verbosity >= (i)) { cout << t << "\n"; cout.flush(); }
#else
#define ENUM_VERB(i,t)
#endif

// private functions
  void maxliteral_update(LPBIState&, int);
  int push_literals_tofix(Coefficient sumci, Coefficient rhs, int iid);
  int fix_lit(int lit);
  int fix_lits(OrdInt& lits);
  int fix_stack();
  
// heuristic dependant stuff
  int select_literal();
  int heuristic;
  int calc_lit_values(double(* val)(Coefficient, Coefficient, Coefficient));
  int heuristic0();
  int val_heuristic_sum(double(* val)(Coefficient, Coefficient, Coefficient));
  int val_heuristic_min(double(* val)(Coefficient, Coefficient, Coefficient));
  int select_maximal_literal_sum(void );
  int select_maximal_literal_min(void );
  int count_fixed_literals(ProductNode* pptr, ProductNode* fixed_pptr, Coefficient sumci, Coefficient rhs);
  int calc_lit_values1(void );
  int calc_lit_values2(void );
  int calc_lit_values3(void );
  int heuristic_max_obj();
  int heuristic4();

  // check whether internal index data structures are correct
  int check_data_structure();
  // show index_datastructures
  int show_data_structure(ostream& out=cout);

  Atoms* atoms;
  // for writing literals
  int plit(int lit, ostream& out = cout)
  {
    if (atoms == 0)
      out << lit;
    else
      {
	if (lit < 0)
	  out << "~" << (*atoms)(-lit);
	else
	  out << (*atoms)(lit);
      }
    return 0;
  }

friend class PBCS;
};

inline int  Enumerate::set_verbosity(int v) { return verbosity = v; }
inline int  Enumerate::get_verbosity() { return verbosity; }
inline void Enumerate::stat_reset() { nodes = fixedlits = largestenlargedETSindex = 0; }
inline int  Enumerate::stat_nodes() { return nodes; }
inline int  Enumerate::stat_maxdepth() { return largestenlargedETSindex; }
inline int  Enumerate::stat_fixedlits() { return fixedlits; }
inline int  Enumerate::set_heuristic(int h) { return heuristic = h; }
inline int  Enumerate::get_heuristic() { return heuristic; }

int coefficient_show_fixed(Coefficient coefficient, int length);

#endif
