#include "Enumerate.h"
#include "timing.h"

// functions for optimizing wrt. current set of LPBIneqs
Coefficient Enumerate::optimize(Products objf,int& consistent,OrdInt& sol, int max, int lb)
{
  if (objf.empty()) // nothing to optimize; just solve
    {
      switch (init_enumeration()) {
      case UNSAT:
	consistent = 0;
	return max ? objf.get_lowerbound()-1 : objf.get_upperbound()+1;
      case SAT:
	consistent = 1;
	return max ? objf.get_upperbound() : objf.get_lowerbound();
      }
      if (solve() == SAT)
	{
	  consistent = 1;
	  sol = last_solution.to_OrdInt();
	  return max ? objf.get_upperbound() : objf.get_lowerbound();
	}
      else
	{
	  consistent = 0;
	  return max ? objf.get_lowerbound()-1 : objf.get_upperbound()+1;
	}
    }
  start_time = used_time();
  Coefficient constant = objf.get_constant();
  Coefficient sumcoeffs = objf.get_sumcoeffs();
  do_maximize = max;
  objf_sumcoeffs = sumcoeffs;
  objf_constant = constant;

  // assure that we maximize an objf where the constant is 0
  // and all ci are positive
  objf.set_constant(0);
  if (max == 0)
    {
      objf.mult(-1);
      objf.set_constant(0);
    }
  // build objective function inequality
#if 1
  // do init faster
  LPBIneq lpbi(objf);
  local_maximum = lpbi.rhs = 0; // do not normalize objfineq
#else
  LPBIneq lpbi;
  objf.first();
  while(!objf.last())
    {
      Product& prod = objf.next();
      lpbi.add(prod.coefficient,prod.literal);      
    }
  local_maximum = lpbi.rhs = 0; // do not normalize objfineq
#endif
  // get unused iid for for objective function
  objective_function_iid = lpbiref.max()+1;
  add(objective_function_iid,lpbi);
  switch (init_enumeration()) {
  case UNSAT:
    consistent = 0;
    return -1+objf_constant;
  case SAT:
    consistent = 1;
    return objf_sumcoeffs + objf_constant;
  }
  if (lb)
    {
      // it is a tautology at the beginning
      EnumTableStack[0].unlink(objective_function_iid); 
      Coefficient lowerbound = LowerBound(objf);
      del(objective_function_iid);
      local_maximum = lpbi.rhs = lowerbound;
      add(objective_function_iid,lpbi);
      if (init_enumeration() == UNSAT)
	{
	  consistent = 0;
	  return -1+objf_constant;
	}
      if (verbosity >= 1)
	{
	  cout << "Initial   ";
	  if (max == 0)
	    {
	      cout << "Minimum: ";
	      coefficient_show_fixed(sumcoeffs-(local_maximum) + constant,16);
	    }
	  else 
	    {
	      cout << "Maximum: ";
	      coefficient_show_fixed(local_maximum + constant,16);
	    }
	  cout << "\n";
	  cout.flush();
	}
    }
  if (verbosity >= 5)
    {
      litref.Print(cout);
      show_enum_table();
    }
  if (lpbi.rhs == 0)
    {
      // it is a tautology at the beginning
      EnumTableStack[0].unlink(objective_function_iid); 
    }
  ENUM_VERB(3,"Start Optimizing");
  maximize();
  // delete objective function inequality
  del(objective_function_iid);
  if (local_maximum == 0) // constraint set infeasible
    consistent = 0;
  else
    consistent = 1;
  sol = last_solution.to_OrdInt();
  if (max == 0)
    return sumcoeffs-(local_maximum-1) + constant;
  else
    return (local_maximum-1) + constant;
}

/* MAXIMIZATION */

int Enumerate::maximize()
{
  int fixlit = select_literal();

  ENUM_VERB(6, " select_literal " << fixlit)
  // complete enumeration anyway
  EnumTablesave();
  if (maximize(fixlit) == OBJUNSAT)
    {
      EnumTablerestore();
      // is current problem with new objective function constraint still infeasible
      LPBIState& obj_state = EnumTableStack[EnumTableStackIndex].operator[](objective_function_iid);
      if (obj_state.rhs > obj_state.sumci)
	{ // infeasible
	  return OBJUNSAT;
	}
      else
	{
	  // goon
	  ENUM_VERB(6,"OBJUNSAT CATCHED")
	  show_enum_table();	    
	  return maximize(-fixlit);
	  // NOTE: It is possible that some literals must be fixed due to the 
	  // new obj function value. This is currently not detected, but whenever
	  // a literal is fixed neg in the obj function it will be detected.
	  // Thus, I think there are no performance problems (maybe ?!?)
	}
    }
  else // UNSAT 
    {
      EnumTablerestore();
      return maximize(-fixlit);
    }
}

int Enumerate::maximize(int fixlit)
{
  nodes++;
  if (verbosity >= 4)
    if (nodes % 500000 == 0)
      {
	cout << "NodeLog : ";
	int_show_fixed(nodes/1000000,6);
	if (nodes % 1000000 == 0)
	  cout << "  ";
	else
	  cout << ".5";
	cout << " million nodes in ";
	double_show_fixed(used_time()-start_time,10);
	cout << " seconds \n";
	cout.flush();
      }
  int res = fix_lit(fixlit);
  if (res == UNSAT)
    return UNSAT;
  else if (res == OPEN)
    return maximize();
  else // SAT
    { // update objective function value and tables      
      // First, get value of obj. function
      Products& pobjf = lpbiref[objective_function_iid];
      int i = 0;
      Coefficient value = 0;
      // first, store the last solution
      last_solution.clear();
      for(i=0;i<=FixedStackIndex;i++)
	last_solution.push(FixedStack[i]);
      // the sum over all coefficients whose literals are bound to 1 or unbound
      ProductNode* pptr = pobjf.head->next;
      while (pptr)
        {
          if (LitValue[pptr->product.literal] != -1) 
            {
              value += pptr->product.coefficient;
	      last_solution.push(pptr->product.literal); // newly fixed ones also
            }
          pptr = pptr->next;
        }
      Coefficient improvement = value - (local_maximum-1);
      local_maximum = value + 1; // try to reach that one!
            
      Coefficient lmax = do_maximize ? value + objf_constant : 
                                       objf_sumcoeffs-value + objf_constant;
      if (verbosity >= 3)
	{
	  cout << "New Local ";
	  if (do_maximize)
	    cout << "Maximum: ";
	  else
	    cout << "Minimum: ";
	  coefficient_show_fixed(lmax,16);
	  cout << "    ";
	  int_show_fixed(nodes,8);cout << " /";
	  int_show_fixed(fixedlits,8);cout << " /";
	  double_show_fixed(used_time()-start_time,10);
	  cout << "\n"; 
	  cout.flush();
	  if (verbosity >= 4)
	    {
	      cout << "Solution : ";
	      OrdInt ls = last_solution.to_OrdInt();
	      ls.first();
	      while(!ls.last())
		{
		  int literal = ls.next();
		  if (literal > 0)
		    cout << (*atoms)(literal) << " ";
		  else
		    cout << "~" << (*atoms)(-literal) << " ";  
		}
	      cout << "\n";cout.flush();
	    }
	}
      if (improvement <= 0)
	{
	  cerr << "Fatal Error : No improvement while maximizing\n";
	  exit(-1);
	}
      // update EnumTables; link in if not satisfied
      Coefficient sumci = objf_sumcoeffs;
      Coefficient rhs = local_maximum;
      int start_next_update = 0; // which literals additionally to fix
      for(i=0;i<=(EnumTableStackIndex-1);i++) // current EnumTable is fathomed!!
	{
	  RArrayList<LPBIState>& table = EnumTableStack[i];
	  LPBIState& objf_state = table[objective_function_iid];
	  if ( table.active_elem(objective_function_iid)) // it is active 
	    objf_state.rhs += improvement; // just update
	  else // it is unactive (unlinked)
	    { // check whether it becomes active again

	      // Things get complicated.
	      // We have to compute all entries from scratch, since they are
	      // not necessarily correct (eliminated entries are not updated (of course))
	      // We retrieve all fixed literals of this state by looking at the
	      // FixedStack at FixedStackChoicepoints[i] 
	      // The literals are incrementally fixed from scratch (efficiency). 
	      // Hence starting from StackBottom is essential.

	      ENUM_VERB(6,"Restoring[" << i << "] sumci = " << sumci << " rhs = " << rhs);
	      if (rhs > 0)
		{
		  // for all fixed literals	that are not yet fixed      
		  for (int j=start_next_update;j <= FixedStackChoicepoints[i];j++)
		    {
		      Coefficient coeff = lpbiref[objective_function_iid].member(FixedStack[j]);
		      if (coeff > 0) // positive fixing
			{
			  sumci -= coeff;
			  rhs -= coeff;
			  if (rhs <= 0)
			    break; // no need to further update; is still unlinked  
			}
		      else if (coeff < 0) // negative fixing
			{
			  sumci -= -coeff;
			  if (sumci < rhs)
			    break; // no need to further update; is still unsatisfiable
			}
		      start_next_update = j+1;
		    }
		}
	      ENUM_VERB(6,"sumci " << sumci << " rhs " << rhs);
	      // Now I have the current values of this node for the objf
	      if (rhs > 0) // we need to relink it in
		{
		  // get maxliteral and maxci
		  Coefficient maxci=0;
		  int maxliteral=0;
		  if (sumci >= rhs) // satisfiable -> hence calc values
		    {
		      ProductNode* hpptr = pobjf.head->next;
		      int sl;
		      int j;
                      while(hpptr)
                        {
                          sl = hpptr->product.literal;
			  int isfixed = 0;
			  for (j=0;j <= FixedStackChoicepoints[i];j++)
			    {
			      int lit = FixedStack[j];
			      if ((lit == sl) || (lit == -sl))
				{
				  isfixed = 1;
				  break; // is fixed			  
				}
			    }
			  if (!isfixed)
			    {			  
			      maxci = hpptr->product.coefficient;
			      maxliteral = sl;
			      break;
			    }
                          hpptr = hpptr->next;
			}
		      if (maxliteral == 0)
			{
			  cerr << " LPBIneqSet : maximize : Fatal Error relink objf inconsistent \n";
			  exit(-1);
			}
		    }
		  table[objective_function_iid].sumci = sumci;
		  table[objective_function_iid].rhs = rhs;
		  table[objective_function_iid].maxliteral = maxliteral;
		  table[objective_function_iid].maxci = maxci;
		  table.insert_front(objective_function_iid);
		}
	    }
	}
      return OBJUNSAT;
    }
}



