#include "LPBIneq.h"

LPBIneq::LPBIneq(Products& p, Coefficient r) : products(p), rhs(r)
{
  normalize();
}

LPBIneq::LPBIneq(Products& p) : products(p) , rhs (0)
{
  // no normalization; only used in EnumerateOpt to speed up init
}

int LPBIneq::initialize(SimpleString& s,Atoms* atoms) 
{ 
  int ret = products.initialize(s,atoms);
  rhs = -products.constant;
  products.constant = 0;
  normalize();
  return ret;
}  

void LPBIneq::add(LPBIneq& lpbi)
{
  ProductNode* h = lpbi.products.head->next;

  while(h)
    {
      products.add(h->product.coefficient,h->product.literal);
      h = h->next;
    }
  rhs += lpbi.rhs;
  rhs -= products.constant;
  products.constant = 0;
  normalize();
}

void LPBIneq::fix(int literal)
{
  products.fix(literal);
  rhs -= products.constant;
  products.constant = 0;
  normalize();
}

void LPBIneq::equate(int lit1, int lit2)
{
  if (products.equate(lit1,lit2)) // only if equation changed something
    {
      rhs -= products.constant;
      products.constant = 0;
      normalize();
    }
}

void LPBIneq::normalize()
{
  rhs -= products.constant;
  products.constant = 0;

  ProductNode* h = products.head->next;
  
  // First, coefficient reduction
  while(h)
    {
      if (h->product.coefficient > rhs)
	{
	  products.sumcoeffs = products.sumcoeffs - (h->product.coefficient-rhs);
	  h->product.coefficient = rhs;
	}
      else
	break;
      h = h->next;
    }
  if (is_tautology()) // is tautology
    {
      products.clear();
      rhs = 0;
    }
  else if (is_contradiction()) // is contradiction
    {
      products.clear();
      rhs = 1;
    }
  else
    gcd_divide();
}

Coefficient my_gcd(Coefficient i, Coefficient j)
{
  if (i == 0)
    return j;
  if (j == 0)
    return i;
  return my_gcd(j%i,i);
}

void LPBIneq::gcd_divide()
{
  ProductNode* h = products.head->next;
  while(h->next)
    h = h->next;
  Coefficient smallest_coefficient = h->product.coefficient;
  if (smallest_coefficient == 1)
    return;
  h = products.head->next;
  Coefficient coeffs_gcd = gcd_coefficient(h,smallest_coefficient);
  if (coeffs_gcd != 1)
    {
//      cout << "found gcd " << coeffs_gcd << " for " << *this << "\n";
      h = products.head->next;
      while(h)
	{
	  h->product.coefficient /= coeffs_gcd;
	  h = h->next;
	}
      products.sumcoeffs /= coeffs_gcd;
      if (my_gcd(rhs,coeffs_gcd) == coeffs_gcd)
	rhs /= coeffs_gcd;
      else
	rhs = (rhs+coeffs_gcd)/coeffs_gcd;
    }
}

Coefficient LPBIneq::gcd_coefficient(ProductNode *h, Coefficient gcd_so_far)
{
  if (h == 0)
    return gcd_so_far;
  return gcd_coefficient(h->next,my_gcd(h->product.coefficient,gcd_so_far));
}

int LPBIneq::fixed_literal()
{
  if (!products.empty())
    {
      products.first();
      Product& product = products.next();
      if (products.sumcoeffs - product.coefficient < rhs)
	return product.literal;
      else
	return 0;
    }
  return 0;
}

int LPBIneq::fix_is_tautology(int literal) 
{
  Coefficient coefficient = member(literal);
  if (coefficient >= rhs)
    return 1;
  else
    return 0;
}

int LPBIneq::is_tautology()
{
  if (rhs <= 0)
    return 1;
  else
    return 0;
}

int LPBIneq::is_contradiction()
{
  if (products.sumcoeffs < rhs)
    return 1;
  else
    return 0;
}

// can be improved when index data in LPBIneqSet is available
int LPBIneq::simply_implies(LPBIneq& lpbi)
{
  // ciLi + cL >= d simply implies ci'Li + c'L' >= d' if
  // ci >= ci' : cL >= d - ci simply implies c'L' >= d'- ci'
  // ci < ci'  : cL >= d simply implies c'L' >= d'
  // a tautology does not imply a non tautology
  Coefficient thisrhs = rhs;
  Coefficient lpbicoefficient = 0;
  ProductNode* actual = products.head->next;
  while(actual)
    {
      lpbicoefficient = lpbi.products.member(actual->product.literal);

      if (lpbicoefficient <= 0)
	thisrhs -= actual->product.coefficient; // a reduction eliminating this product
      else
	{
	  if (actual->product.coefficient >= lpbicoefficient)
	    thisrhs -= actual->product.coefficient - lpbicoefficient; // reduce by difference
	  // else forget it
	}
      if (thisrhs < lpbi.rhs)
	return 0;
      actual = actual->next;
    }
//  cout << *this << " implies " << lpbi << "\n";
  return 1;
}

int LPBIneq::check_data_structure()
{
  Coefficient sc = products.sumcoeffs;
  
  products.first();
  while(!products.last())
    sc -= products.next().coefficient;
  if (sc != 0)
    {
      cerr << "LPBIneq : check_data_structure : sumcoeffs not correct\n";
      return 1;
    }  
  return 0;
}

int LPBIneq::Read(istream& in,Atoms* atoms)
{
  clear();
  products.Read(in,atoms);
  rhs -= products.constant;
  products.constant = 0;
  normalize();
  return 0;
}

void LPBIneq::Print(ostream& out,Atoms* atoms)
{
  products.Print(out,atoms);
  out << " >= " << rhs;
}
