#include "PBCS.h"
#include "timing.h"

PBCS::PBCS()
{
  atoms = 0;
  verbosity = 3;
  is_unsat = 0;
  max_cut_nodes = 0;
  no_covers = 0;
  enum_dirty = 1;
  enum_state = 0;  
}

PBCS::PBCS(const PBCS& pbcs)
{
  S = pbcs.S;
  Eq = pbcs.Eq;
  Fixed = pbcs.Fixed;
  Enum = pbcs.Enum;  
  atoms = pbcs.atoms;
  is_unsat = pbcs.is_unsat;
  max_cut_nodes = pbcs.max_cut_nodes;
  no_covers = 0;
  enum_dirty = 1;
  enum_state = 0;
}

PBCS& PBCS::operator=(const PBCS& pbcs)
{
  S = pbcs.S;
  Eq = pbcs.Eq;
  Fixed = pbcs.Fixed;
  Enum = pbcs.Enum;
  atoms = pbcs.atoms;
  is_unsat = pbcs.is_unsat;
  max_cut_nodes = pbcs.max_cut_nodes;
  no_covers = 0;
  enum_dirty = 1;
  enum_state = 0;
  return *this;
}

PBCS::~PBCS()
{
  // nothing special to do
}

void PBCS::clear()
{
  S.clear();
  Eq.clear();
  Fixed.clear();
  Enum.clear();
  is_unsat = 0;
  enum_dirty = 1;
  enum_state = 0;
}

int PBCS::apply_fix_eq(LPBIneq& lpbi)
{
  // apply equations 
  Eq.first();
  while(!Eq.last())
    {
      int l = Eq.next();
      int r = Eq(l);
      lpbi.equate(r,l); // unnatural ordering significant!
    }
  // apply fixings
  Fixed.first();
  while(!Fixed.last())
    {
      lpbi.fix(Fixed.next());
    }
  return 0;
}

int PBCS::add_prepro(LPBIneq& lpbi)
{
  if (lpbi.is_contradiction())
    {
      set_unsat();
      return -1;
    }
  if (lpbi.is_tautology())
    return 1;  
  return 0;
}

int PBCS::add(LPBIneq& origlpbi, int internal)
{
  if (is_unsatisfiable()) // ignore ineqs if constraint set is already unsat
    return 0;
  LPBIneq lpbi(origlpbi); // don't change param
  if (!internal)
    apply_fix_eq(lpbi);
  int ret = add_prepro(lpbi);
  if ((ret == -1) || (ret == 1)) // unsat or tautology
    return 0;
  // else ok; an ineq to add
  int iid = S.add(lpbi);
  Enum.add(iid,lpbi);
  enum_dirty = 1;
  return iid;
}

int PBCS::add_delete_implied(LPBIneq& origlpbi, int internal)
{
  LPBIneq lpbi = origlpbi; // don't change param
  if (!internal)
    apply_fix_eq(lpbi);
  switch(add_prepro(lpbi)) {
  case -1 : // unsat
  case 1  : // tautology
    return 0;
  }
  // check whether ineq is implied by an ineq in S
  if (S.simply_implies(lpbi))
    return 0; // adding discarded, since implied
  // ok; an ineq to add
  // delete all ineqs that are simply implied by lpbi
  delete_implied_by(lpbi);
  int iid = S.add(lpbi);
  Enum.add(iid,lpbi);    
  enum_dirty = 1;
  return iid;
}

int PBCS::del_lpbi(int iid)
{
  S.del(iid);
  Enum.del(iid);
  enum_dirty = 1;
  return 0;
}


int PBCS::delete_implied_by(LPBIneq& lpbi)
{
  OrdInt todel;
  S.alliids.first();
  while(!S.alliids.last())
    {
      int iid = S.alliids.next();
      if (lpbi.simply_implies(S.Pool[iid]))
	todel += iid;
    }
  todel.first();
  while(!todel.last())
    del_lpbi(todel.next());
  return todel.length();
}

int PBCS::delete_implied()
{
  OrdInt todel;
  OrdInt alliids = S.alliids;
  OrdInt checkiids = S.alliids;
  checkiids.first();
  while(!checkiids.last())
    {
      int check_iid = checkiids.next();
      LPBIneq& check_lpbi = S.Pool[check_iid];
      alliids.first();
      while(!alliids.last())
	{
	  int iid = alliids.next();	  
	  if ((iid != check_iid) && (!todel.member(iid)))
	    { 
	      if (S.Pool[iid].simply_implies(check_lpbi))
		todel += check_iid;
	    }
	}
    }
  todel.first();
  while(!todel.last())
    del_lpbi(todel.next());
  return todel.length();
}

// BUGGY! must not consider ineq to eliminate while calc. lower bound
int PBCS::delete_redundant()
{
  if (init_enumeration())
    return 0;
  OrdInt todel;
  OrdInt alliids = S.alliids;
  alliids.first();
  while(!alliids.last())
    {
      int iid = alliids.next();
      LPBIneq& lpbi = S.Pool[iid];
      Coefficient lower_bound = Enum.LowerBound(lpbi.products);
      if (lower_bound >= lpbi.rhs)
	{
	  todel += iid;
	  cout << "Bound: " << lower_bound << lpbi << "\n";
	}
    }
  todel.first();
  while(!todel.last())
    del_lpbi(todel.next());
  return todel.length();
}

OrdInt PBCS::single_vars()
{
  OrdInt sv;
  if (is_unsatisfiable())
    return sv;
  for (int i=1;i<=Enum.litref.upper();i++)
    {
      if (Enum.litref.var_no_occurs(i) == 1)
	{
	  if (Enum.litref.lit_no_occurs(i) == 1)
	    sv += i;
	  else 
	    sv += -i;
	}
    }       
  return sv;
}

int PBCS::coversolve()
{
  int clen;
  int res;
  while((res = addcover(clen)) == 0)
    ;
  return res;
}

Coefficient PBCS::optimize(Products objf, int& consistent, OrdInt& sol, int max, int lb)
{
  // apply equations 
  Eq.first();
  while(!Eq.last())
    {
      int l = Eq.next();
      int r = Eq(l);
      objf.equate(r,l); // unnatural ordering significant!
    }
  // apply fixings
  Fixed.first();
  while(!Fixed.last())
    {
      objf.fix(Fixed.next());
    }
  // call enumeration optimization
  return Enum.optimize(objf, consistent, sol, max, lb);
}

int int_show_fixed(int i,int length); // in EnumerateOpt.C

Coefficient PBCS::coveroptimize(Products objf, int& consistent, OrdInt& sol, int max,int)
{
  double start_time = used_time();
  // apply equations 
  Eq.first();
  while(!Eq.last())
    {
      int l = Eq.next();
      int r = Eq(l);
      objf.equate(r,l); // unnatural ordering significant!
    }
  // apply fixings
  Fixed.first();
  while(!Fixed.last())
    {
      objf.fix(Fixed.next());
    }
  Coefficient constant = objf.constant;
  objf.constant = 0;
  // do cover optimization
  // First; get a solution
  int clen;
  int res;
  while((res = addcover(clen)) == 0)
    ;
  if (res == -1) // problem is unsatisfiable
    {
      sol.clear();
      consistent = 0;
      return objf.constant -1;
    }
  consistent = 1;
  // hence res = 1 and we have a solution found
  sol = get_last_solution();
  Coefficient last_optimum,remaining;
  do 
    {
      // always just having a solution here!
      OrdInt ls = Enum.last_solution.to_OrdInt();
      objf.get_value(ls,last_optimum,remaining);
      if (verbosity >= 3)
	{
	  if (verbosity >= 4)
	    cout << "\n";
	  cout << "New Local ";
	  if (max)
	    {
	      cout << "Maximum: ";
	      coefficient_show_fixed(last_optimum+remaining+constant,16);
	    }
	  else
	    {
	      cout << "Minimum: ";
	      coefficient_show_fixed(last_optimum+constant,16);
	    }
	  cout << "    ";
	  int_show_fixed(no_covers,8);cout << " /";
	  int_show_fixed(Enum.nodes,8);cout << " /";
	  int_show_fixed(Enum.fixedlits,8);cout << " /";
	  double_show_fixed(used_time()-start_time,10);
	  cout << "\n"; 
	  cout.flush();
	}     
      if (max == 0)
	{
	  Products objfcopy(objf);
	  objfcopy.mult(-1);
	  LPBIneq lpbi(objfcopy,-(last_optimum-1));
//	  lpbi.Print(cout,atoms);cout << "\n";
	  add_delete_implied(lpbi); // add new bound
	}
      else
	{
	  LPBIneq lpbi(objf,last_optimum+remaining+1);
//	  lpbi.Print(cout,atoms);cout << "\n";
	  add_delete_implied(lpbi); // add new bound
	}
      // Cut generation until a solution is found or unsatisfiability is detected
      while((res = addcover(clen)) == 0)
	;
      if (res == 1)
	sol = get_last_solution();
    } while (res != -1); // as long as satisfiable
  return max ? last_optimum + remaining + constant : last_optimum + constant;
}

int PBCS::init_enumeration()
{
  if (is_unsatisfiable())
    return -1;
  if (!enum_dirty && 0) // disable for the moment (buggy)
    enum_state = Enum.reset_enumeration(); // a warm reset 
  else
    enum_state = Enum.init_enumeration();
  enum_dirty = 0;
  if (enum_state == Enum.UNSAT) 
    {
      set_unsat();
      return -1;
    }
  else if (enum_state == Enum.SAT)
    return 1;
  else if (enum_state == Enum.OPEN)
    return 0;
  cerr << "unknown return value from Enum.init_enumeration()\n";
  return -1;
}

int PBCS::set_unsat()
{
  clear();
  is_unsat = 1;
  enum_dirty = 1;
  return 1;
}

int PBCS::datastructure_corrupted()
{
  int res;

  S.first();
  while(!S.last())
    {
      LPBIneq& lpbi = S.next();
      if ((res = lpbi.check_data_structure()))
	return res;
    }
  if ((res = Enum.check_data_structure()))
    return res;
  return 0;
}


