#ifndef PBCS_h
#define PBCS_h

#include "LPBIneqSet.h"
#include "Equations.h"
#include "Enumerate.h"
#include "Atoms.h"

class PBCS {
public:
  PBCS();
  PBCS(const PBCS& );
  PBCS& operator=(const PBCS&);
  ~PBCS();

  void clear();
  int apply_fix_eq(LPBIneq& lpbi);
  int add(LPBIneq& lpbi,int internal=0); // default applies fixing and equations
  int add_delete_implied(LPBIneq& lpbi, int internal=0);

  int del_lpbi(int iid);
  LPBIneq get_lpbi(int iid);
  OrdInt get_lpbi_iids(); 
  int is_empty() { return get_lpbi_iids().empty(); }

  // eliminates inequalities simply implied by a single inequality
  // returns number of eliminated inequalities (quadratic complexity)
  int delete_implied(); 
  // eliminates all constraints simply implied by lpbi
  int delete_implied_by(LPBIneq& lpbi);
  // eliminates redundant inequalities (via lowerbound)
  // returns number of eliminated inequalities
  int delete_redundant();  // still buggy

  OrdInt single_vars(); // all variables that occur just once
  OrdInt get_fixed(); // returns all fixed literals
  Equations get_equations();


  int coversolve(); // solves problem by cover generation -> -1 UNSAT; 1 SAT

  // optimizing wrt. current set of constraints
  Coefficient optimize(Products objf, int& consistent, OrdInt& sol, int max=1, int lb=0);
  // optimize via a pure logic cut algorithm
  // Attention! contraint set is inconsistent afterwards
  Coefficient coveroptimize(Products objf, int& consistent, OrdInt& sol, int max=1,int lb=0);

  // options settings and gettings
  int set_enum_heuristic(int h) { return Enum.set_heuristic(h); }
  int get_enum_heuristic() { return Enum.get_heuristic(); }
  int set_enum_verbosity(int v) { return Enum.set_verbosity(v); }
  int get_enum_verbosity() { return Enum.get_verbosity(); }
  int set_verbosity(int v) { return verbosity = v; }
  int get_verbosity() { return verbosity; }

  int set_atoms(Atoms* global_atoms);
  // stuff for statisticscontrol
  void enum_stat_reset() { Enum.stat_reset(); }

// the data (later there will also be an ExtClauseSet :-)
  LPBIneqSet S;    // the set of linear pseudo-Boolean inequalities
  Equations Eq;    // the equations between literals;
  OrdInt Fixed;    // the fixed literals;
  Enumerate Enum;  // the enumeration datastructure
  int init_enumeration(); // initializes enumeration
  int enum_dirty;  // checks whether enumeration needs to be reinitialized
  int enum_state;  // if not, what is the current state

  int is_unsat;    // is_unsat == 1 iff the constraint set is unsatisfiable
  int is_unsatisfiable() { return is_unsat; }
  
  int datastructure_corrupted(); // check whether all is ok

#include "PBCSfixing.h"
#include "PBCSequations.h"
#include "PBCScoeffred.h"
#include "PBCScut.h"

  int verbosity;
private:
// for printing an Atoms structure
  Atoms* atoms;
  int plit(int lit,ostream& out = cout);
  int add_prepro(LPBIneq& lpbi); // preprocessing before adding
  int set_unsat(); // sets is_unsat and clears all
};

inline LPBIneq PBCS::get_lpbi(int iid)
{
  return S.get_lpbi(iid);
}

inline OrdInt PBCS::get_lpbi_iids()
{
  return S.get_iids();
}

inline OrdInt PBCS::get_fixed()
{
  return Fixed;
}

inline Equations PBCS::get_equations()
{
  return Eq;
}

inline int PBCS::set_atoms(Atoms* global_atoms)
{
  atoms = global_atoms;
  Enum.atoms = global_atoms;
  return 0;
}

inline int PBCS::plit(int lit,ostream& out)
{
  if (atoms == 0)
    out << lit;
  else
    {
      if (lit < 0)
	out << "~" << (*atoms)(-lit);
      else
	out << (*atoms)(lit);
    }
  return 0;
}

#endif
