/* opbdp.C
 * Optimizing Pseudo-Boolean Davis-Putnam
 * Author : Peter Barth
 * Date   : 17.5.95
 */

#include "streaminclude.h"
#include "fstream.h"
#include "timing.h"
#include "PBCS.h"
#include "MGetOpt.h"
#include "opbdp_read.h"
#include "readnonlinear.h"
#include "lp_write.h"


void header(char* argv0)
{
  cout << "opbdp 1.1 #1  Copyright (C) 1995-98 Peter Barth <barth@mpi-sb.mpg.de>\n";
  cout << "              Max-Planck-Institut f. Informatik\n";
  cout << "type " << argv0 << " -H for help\n";
}

void usage(char* argv0)
{
  header(argv0);
  cout << "Options:\n";
  cout << "  -vn    verbosity\n";
  cout << "    n=0: show nothing; return value is optimum \n";
  cout << "    n=1: show optimal value + init-information \n";
  cout << "    n=3: show each local optima <default> \n";
  cout << "    n=4: a short notice all 500000 nodes \n";
  cout << "  -hn    variable selection heuristics\n";
  cout << "    n=0: no heuristics at all; first unfixed literal positive \n";
  cout << "    n=1: adapted two-sided Jeroslow-Wang heuristics <default> \n";
  cout << "    n=2: similar to h1; probably not exceeding double range \n";
  cout << "    n=3: fixing a literal yields maximal further fixings in the next step\n";
  cout << "    n=4: randomly selects a literal (rand)\n";
  cout << "    n=1x:  min(H[Li],H[-Li]) is maximal for -hx \n";
  cout << "           (h1-3: H[Li] + H[-Li] is maximal) \n";
  cout << "    n=1xx: literal with greatest coefficient in objective. function; \n";
  cout << "           -hxx if none left\n";
  cout << "  -px*   preprocessing \n";
  cout << "    x=f: search for fixed literals (if L=1 -> 1=0 then L = 0)\n";
  cout << "    x=e: search for equations between literals \n";
  cout << "         If L=1 -> B=1 and L=0 -> B=0then L = 1-B \n";
  cout << "    x=c: coefficient reduction\n";
  cout << "    x=M: coefficient reduction for coefficients that equal the right-hand side\n";
  cout << "    x=F: search for fixed literals implied by a single inequality\n";
  cout << "    Preprocessing is done in the order the option character occur\n";
  cout << "    Multiple preprocessing may be worth while and is possible. Default is x=FM\n";
  cout << "  -wx<file> writes preprocessed problem to <file>\n";
  cout << "    x=o: format readable for opbdp \n";
  cout << "    x=l: format readable for lp_solve \n";
  cout << "    x=c: format readable for cplex \n";
  cout << "  -b     calculates an initial bound of the objective function \n";
  cout << "  -s     show the solution (all 0-1 variables that are fixed to 1) \n";
  cout << "  -f<name> read from file <name> instead of reading from stdin \n";
  cout << "  -nx    read a nonlinear problem and linearize it \n";
  cout << "         x=f -> Fortet Linearization; x=p -> my linearization method (default) \n";
  cout << "  -i     interactive mode; a simple command line interface to opbdp \n";
  cout << "  -H     this screen\n";
}

// a global atoms table, where all variable names are stored 
// together with their internal representation (a positive number)
Atoms* globalatoms = 0;

int cout_literal_(int literal)
{
  if (literal > 0)
    cout << (*globalatoms)(literal) << " ";
  else
    cout << "~" << (*globalatoms)(-literal) << " ";  
  return 0;
}

#ifdef QUANTIFIED
extern "C" {
quantify_start_recording_data();
quantify_stop_recording_data();
}
#endif

int opbdp_main(int argc, char **argv);
int pbcsia_main(int argc, char **argv);

int main(int argc, char** argv)
{
  int ret = 0;
  int ia = 0;
  int option_char;
  MGetOpt getopt (argc, argv);

  while (((option_char = getopt.optchar()) != EOF) && (ia == 0))
    switch (option_char)
      {
      case 'i' : 
	ret = pbcsia_main(argc, argv);
	ia = 1;
	break;
      default:
	break;
      }
  if (!ia)
    ret = opbdp_main(argc, argv);
  return ret;
}

int opbdp_main(int argc, char **argv)
{
  #ifdef QUANTIFIED
  quantify_stop_recording_data();
  #endif

  MGetOpt getopt (argc, argv);
  int option_char;
  int verbosity = 3;
  int heuristic = 1;
  int from_file= 0;
  int show_solution = 0;
  int max = 1;
  SimpleString filename;
  SimpleString preprostring = "FM";
  int just_write = 0;
  int write_solver = OPBDP;
  SimpleString write_filename;
  int do_ib = 0;
  int nonlinear = 0;

  // Remark: don't use this datastrcutures as global variables.
  // It segmentation faults on clean up.
  // Intuitively, I make the fuzzy C++-compilers responsible...

  // PBCS; a class for Pseudo-Boolean Constraint Solving
  LPBIneqSet X;    // for reading the problem
  PBCS pbcs;       // the set of linear pseudo-Boolean inequalities is in PBCS,
  Products objf;   // the objective function
  Atoms atoms;     // and their Atomstable
  globalatoms = &atoms;
  pbcs.set_atoms(&atoms);
  OrdInt sol;     // the solution

  // timing stuff
  float tmp_time = 0.0;
  float start_time = 0.0;
  float read_time = 0.0;
  float solve_time = 0.0;
  float end_time = 0.0;
  float prepro_time = 0.0;

  // command line option processing
  while ((option_char = getopt.optchar()) != EOF)
    switch (option_char)
      {
      case 'v': verbosity = atoi(getopt.opttext()); break;
      case 'f': filename = getopt.opttext(); from_file = 1; break;
      case 'h': heuristic = atoi(getopt.opttext()); break;
      case 's': show_solution = 1; break;
      case 'p': 
	preprostring = getopt.opttext();
	break;
      case 'w' :
        {
	  just_write = 1;
	  SimpleString s = getopt.opttext();	  
	  if (s.length() < 2)
	    {
	      cerr << "filename missing for -w option\n";
	      exit(-1);
	    }
	  switch (s[0])
	    {
	    case 'o' :
	      write_solver = OPBDP;
	      write_filename = s(1,s.length()-1);
	      break;
	    case 'c' :
	      write_solver = CPLEX;
	      write_filename = s(1,s.length()-1);
	      break;
	    case 'l' :
	      write_solver = LPSOLVE;
	      write_filename = s(1,s.length()-1);
	      break;
	    default :
	      write_solver = OPBDP;
	      write_filename = s(0,s.length()-1);
	      break;
	    }
	}
	break;
      case 'n' :
	{
	  SimpleString s = getopt.opttext();	  
	  if (s.length() < 1)
	    nonlinear = 2;
	  else
	    {
	      switch (s[0])
		{
		case 'f' : 
		  nonlinear = 1;
		  break;
		case 'p' : 
		  nonlinear = 2;
		  break;
		default :
		  {
		    cerr << "unknown linearization method " << s[0] << " requestes \n";
		    exit(-1);
		  }
		}
	    }
	}
	break;
      case 'b':
	do_ib = 1;
        break;
      case 'H': 
      default : usage(argv[0]);
	return 0;
      }
  header(argv[0]);
  pbcs.set_enum_heuristic(heuristic);
  pbcs.set_enum_verbosity(verbosity);
  pbcs.set_verbosity(verbosity);
  start_time = used_time();
  if (from_file)
    {
      if (filename.length() == 0)
	{
	  cerr << "You haven't specified a problem file with -f\n";
	  exit(-1);
	}
      else
	{
	  char charfilename[80];
	  int i;
	  for(i=0;i < filename.length();i++)
	    charfilename[i] = filename[i];
	  charfilename[i] = 0;
	  ifstream in(charfilename);
	  if (!in)
	    {
	      cerr << "Problem with file " << filename << "\n";
	      exit(-1);
	    }
	  if (nonlinear == 0)
	    {
	      if (opbdp_read(atoms,X,objf,in,max))
		{
		  cerr << "Problem reading data\n";
		  exit(-1);
		}
	    }
	  else
	    {
	      if (readnonlinear(in,objf,X,max,atoms,nonlinear-1,verbosity))
		{
		  cerr << "Problem reading nonlinear data\n";
		  exit(-1);
		}		
	    }
	  X.first();
	  while(!X.last())
	    {
	      LPBIneq& lpbi = X.next();
	      pbcs.add(lpbi);
	    }
	}
    }
  else // read it from stdin
    {
      if (nonlinear == 0)
	{
	  if (opbdp_read(atoms,X,objf,cin,max))
	    {
	      cerr << "Problem reading data\n";
	      exit(-1);
	    }
	}
      else
	{
	  if (readnonlinear(cin,objf,X,max,atoms,nonlinear-1,verbosity))
	    {
	      cerr << "Problem reading nonlinear data\n";
	      exit(-1);
	    }		
	}
      X.first();
      while(!X.last())
	{
	  LPBIneq& lpbi = X.next();
	  pbcs.add(lpbi);
	}
    }
  cout.flush();
  read_time = used_time()-start_time;
  tmp_time = used_time();
  pbcs.Enum.stat_reset();

  // preprocessing
  
  int i=0;
  while(i < preprostring.length())
    {
      if (preprostring[i] == 'F')
	{
	  OrdInt fixed_lits = pbcs.fixing();
	  if (!fixed_lits.empty())
	    {
	      int lit;
	      cout << "Fixing : ";
	      fixed_lits.first();
	      while(!fixed_lits.last())
		{
		  objf.fix(lit = fixed_lits.next());
		  cout_literal_(lit);
		}
	      if (pbcs.is_unsatisfiable())
		cout << "\n'fixing' detected inconsistency\n";		
	      else if (fixed_lits.length() == 1)		
		cout << "\n1 Literal has been fixed by 'fixing' \n";
	      else
		cout << "\n" << fixed_lits.length() << " Literals have been fixed by 'fixing' \n";	  
	    }
	}
      else if (preprostring[i] == 'f')
	{
	  OrdInt fixed_lits = pbcs.fixingcheck();
	  if (!fixed_lits.empty())
	    {
	      int lit;
	      cout << "Fixingcheck : ";
	      fixed_lits.first();
	      while(!fixed_lits.last())
		{		  
		  objf.fix(lit = fixed_lits.next());
		  pbcs.fix(lit);
		  cout_literal_(lit);
		}
	      if (pbcs.is_unsatisfiable())
		cout << "\n'fixingcheck' detected inconsistency\n";		
	      else if (fixed_lits.length() == 1)		
		cout << "\n1 Literal has been fixed by 'fixingcheck'\n";
	      else
		cout << "\n" << fixed_lits.length() << " Literals have been fixed by 'fixingcheck' \n";
	      cout.flush();
	    }
	}
      else if (preprostring[i] == 'e')
	{
	  int no_neweqs = pbcs.eq_check();
	  if (no_neweqs != 0)
	    {
	      Equations& eq=pbcs.Eq;
	      eq.first();
	      if (!eq.last())
		{
		  int eqs=0;
		  int l,r;
		  cout << "Implications_equations : ";
		  while(!eq.last())
		    {	
		      l = eq.next();
		      r = eq(l);
		      objf.equate(r,l);
		      cout_literal_(l); cout << "= "; cout_literal_(r);
		      cout.flush();
		      eqs++;
		    }
		  if (no_neweqs == 1)		
		    cout << "\n1 new Equation has been detected and applied\n";
		  else
		    cout << "\n" << no_neweqs << " new Equations have been detected and applied\n";
		  cout.flush();
		}
	    }
	}
      else if (preprostring[i] == 'M')
	{
	  int changed = pbcs.coefficient_reduction_bigM();
	  if (changed > 1)	    
	    cout << "Big M coefficient reduction changed " << changed << " coefficients\n";
	  else if (changed == 1)
	    cout << "Big M coefficient reduction changed " << changed << " coefficient\n";
	  cout.flush();
	}
      else if (preprostring[i] == 'c')
	{
	  int changed = pbcs.coefficient_reduction();
	  if (changed > 1)	    
	    cout << "General coefficient reduction changed " << changed << " coefficients\n";
	  else if (changed == 1)
	    cout << "General coefficient reduction changed " << changed << " coefficients\n";
	  cout.flush();
	}
      else
	{
	  cout << "unknown preprocessing method " << preprostring[i] << "\n";
	}
      i++;
    }
  prepro_time = used_time() - tmp_time;
  cout << "Preprocessingtime (secs) : "; double_show_fixed(prepro_time,2); cout << "\n";

  tmp_time = used_time();
  if (just_write)
    {
      char xfilename[80];
      for(i=0;i < write_filename.length();i++)
	xfilename[i] = write_filename[i];
      xfilename[i] = 0;      
      ofstream lp(xfilename);
      lp_write(lp,pbcs.S,objf,max,&atoms,write_solver);
      end_time = used_time();
      solve_time = end_time-tmp_time;
      cout << "Problem written to file " << xfilename << " in ";
      switch (write_solver)
	{
	case OPBDP   : cout << "OPBDP";  break;
	case CPLEX   : cout << "CPLEX";  break;
	case LPSOLVE : cout << "LPSOLVE";break;
	default : break;
	}
      cout << "-format in " << solve_time << "sec. \n";
      return 0;
    }
  pbcs.Enum.stat_reset();
  int consistent;
  cout << "                               Value        Nodes /    Fixed /       Time \n";
  cout.flush();
  #ifdef QUANTIFIED
  quantify_start_recording_data();
  #endif
  Coefficient value;
  if (pbcs.is_unsatisfiable())
    {
      value = 0;
      consistent = 0;
    }
  else if (pbcs.is_empty())
    {
      value = max ? objf.get_upperbound() : objf.get_lowerbound();
      consistent = 1;
    }
  else
    {
      ////////////////////////////////
      value = pbcs.Enum.optimize(objf,consistent,sol,max,do_ib);
      ////////////////////////////////
    }
  #ifdef QUANTIFIED
  quantify_stop_recording_data();
  #endif

  if (consistent == 0)
    {
      cout << "Constraint Set is unsatisfiable\n";
    }
  else
    if (verbosity >= 1)
      {
	cout << "Global ";
	if (max)
	  cout << "Maximum: ";
	else
	  cout << "Minimum: ";
	cout << "****** " << value << " ******\n";
      }
  end_time = used_time();
  solve_time = end_time-tmp_time;
  if (verbosity >= 1)
    {
      cout << "Explored Nodes: " << pbcs.Enum.stat_nodes();
      cout << "   Max. Depth: " << pbcs.Enum.stat_maxdepth();
      cout << "   Fixed Literals: " << pbcs.Enum.stat_fixedlits() << "\n"; 
      cout << "Time (secs) : init("; double_show_fixed(read_time,2); 
      cout << ") prepro("; double_show_fixed(prepro_time,2);
      cout << ") solve("; double_show_fixed(solve_time,2); cout << ")\n";
      if (show_solution)
	{
	  OrdInt ones;
	  cout << "0-1 Variables fixed to 1 : ";
	  sol.first();
	  while(!sol.last())
	    {
	      int var = sol.next();
	      if (var > 0)
		ones += var;
	    }
	  OrdInt Fixed = pbcs.get_fixed();
	  Fixed.first();
	  while(!Fixed.last())
	    {
	      int flit = Fixed.next();
	      if (flit > 0)
		ones += flit;
	    }
	  // add hidden ones due to equations
	  pbcs.Eq.first();
	  while(!pbcs.Eq.last())
	    {
	      int l = pbcs.Eq.next();
	      int r = pbcs.Eq(l);
	      if (sol.member(r) || Fixed.member(r)) 
		ones += l;
	      else if (sol.member(-r) || Fixed.member(-r))
		ones += l;
	      // just positives are added
	    }
	  ones.first();
	  while(!ones.last())
	    {
	      int var = ones.next();
	      cout << atoms(var) << " ";
	    }
	  cout << "\n";
	}
    }
  if (verbosity == 0)
    return coefficient_to_int(value);
  else
    return 0;
}



